import asyncio
from collections.abc import AsyncGenerator
from collections.abc import Awaitable
from os.path import join
from typing import Optional  # noqa: F401 # Needed until tiangolo/typer#522 is fixed)

from pgbelt.config.models import DbupgradeConfig
from pgbelt.config.remote import resolve_remote_config
from pgbelt.util import get_logger
from pgbelt.util.asyncfuncs import isdir
from pgbelt.util.asyncfuncs import isfile
from pgbelt.util.asyncfuncs import listdir


def get_config(
    db: str, dc: str, skip_src: bool = False, skip_dst: bool = False
) -> Optional[DbupgradeConfig]:
    """
    Get a configuration for one database pair synchronously.
    """
    config = asyncio.run(get_config_async(db, dc, skip_src, skip_dst))

    if config is None:
        exit(1)

    return config


async def get_config_async(
    db: str, dc: str, skip_src: bool = False, skip_dst: bool = False
) -> Optional[DbupgradeConfig]:
    """
    Get configuration for one database pair asynchronously. Always prefers
    locally cached configuration but attempts to resolve any uncached configuration
    if it is required. Locally cached config never expires, so it may become stale.
    """
    logger = get_logger(db, dc, "config")
    logger.info("Getting configuration...")
    config = await DbupgradeConfig.load(db, dc)

    if config is None or (config.src is None and config.dst is None):
        logger.info("Resolving remote configuration...")
        config = await resolve_remote_config(db, dc, skip_src, skip_dst)
    elif config.src is None and not skip_src:
        logger.info("Cached config did not include source info, resolving...")
        src_config = await resolve_remote_config(db, dc, skip_dst=True)
        if src_config is None or src_config.src is None:
            logger.critical("Could not resolve missing source info!")
            return None
        config.src = src_config.src
    elif config.dst is None and not skip_dst:
        logger.info("Cached config did not include target info, resolving...")
        dst_config = await resolve_remote_config(db, dc, skip_src=True)
        if dst_config is None or dst_config.dst is None:
            logger.critical("Could not resolve missing target info!")
            return None
        config.dst = dst_config.dst
    else:
        return config

    if config is None or (config.src is None and config.dst is None):
        logger.critical("No configuration could be retrieved")
    else:
        await config.save()

    return config


async def find_available_configs(confdir: str, dc: str) -> set[str]:
    """
    Search for all the database configs in a datacenter directory within a config directory.
    """
    result = set()
    dc_dir = join(confdir, dc)
    if not await isdir(dc_dir):
        return set()
    for db in await listdir(dc_dir):
        if await isdir(join(dc_dir, db)) and await isfile(
            join(dc_dir, db, "config.json")
        ):
            result.add(db)
    return result


async def get_all_configs_async(
    dc: str, skip_src: bool = False, skip_dst: bool = False
) -> AsyncGenerator[Awaitable[Optional[DbupgradeConfig]], None]:
    """
    A generator that produces Awaitables that resolve to DbupgradeConfigs or None

    Will produce all possible configs in the given dc whether they are remote
    or already resolved and cached.
    """
    logger = get_logger("all", dc, "config")
    logger.info("Getting all available configurations...")
    remote, local = await asyncio.gather(
        find_available_configs("remote-configs", dc),
        find_available_configs("configs", dc),
    )

    for conf in asyncio.as_completed(
        [
            get_config_async(db, dc, skip_src=skip_src, skip_dst=skip_dst)
            for db in remote | local
        ]
    ):
        yield conf



import socket
from asyncio import open_connection
from asyncio import run
from asyncio import TimeoutError
from asyncio import wait_for
from collections.abc import Awaitable
from logging import Logger

from asyncpg import create_pool
from pgbelt.cmd.helpers import run_with_configs
from pgbelt.config.config import get_config
from pgbelt.config.models import DbupgradeConfig
from pgbelt.util.logs import get_logger
from pgbelt.util.postgres import analyze_table_pkeys
from tabulate import tabulate
from typer import echo
from typer import Option
from typer import style


def src_dsn(
    db: str,
    dc: str,
    owner: bool = Option(False, help="Use the owner credentials"),
    pglogical: bool = Option(False, help="Use the pglogical credentials."),
) -> None:
    """
    Print a dsn to stdout that you can use to connect to the source db:
    psql "$(dbup src-dsn scribble prod-use1-pg-1)"

    Pass --owner to log in as the owner or --pglogical to log in as pglogical.
    """
    conf = get_config(db, dc, skip_dst=True)
    echo(
        conf.src.owner_dsn
        if owner
        else conf.src.pglogical_dsn if pglogical else conf.src.root_dsn
    )


def dst_dsn(
    db: str,
    dc: str,
    owner: bool = Option(False, help="Use the owner credentials"),
    pglogical: bool = Option(False, help="Use the pglogical credentials."),
) -> None:
    """
    Print a dsn to stdout that you can use to connect to the destination db:
    psql "$(dbup dst-dsn scribble prod-use1-pg-1)"

    Pass --owner to log in as the owner or --pglogical to log in as pglogical.
    """
    conf = get_config(db, dc, skip_src=True)
    echo(
        conf.dst.owner_dsn
        if owner
        else conf.dst.pglogical_dsn if pglogical else conf.dst.root_dsn
    )


async def _check_pkeys(
    conf: DbupgradeConfig, logger: Logger
) -> tuple[list[str], list[str]]:
    async with create_pool(conf.src.root_uri, min_size=1) as pool:
        pkey_tables, no_pkey_tables, _ = await analyze_table_pkeys(
            pool, conf.schema_name, logger
        )
    return pkey_tables, no_pkey_tables


def check_pkeys(db: str, dc: str) -> None:
    """
    Print out lists of tables with and without primary keys
    """
    conf = get_config(db, dc, skip_src=True)
    logger = get_logger(db, dc, "convenience.src")
    pkeys, no_pkeys = run(_check_pkeys(conf, logger))
    echo(
        f"""Analyzed table pkeys for {db} in {dc}:
        has pkey: {pkeys}
        no pkey: {no_pkeys}
        """
    )


async def _print_connectivity_results(results: list[dict]):
    """
    For a list of databases in a datacenter, show a table of established connections.

    Also exit(1) if ANY connections failed.
    """

    table = [
        [
            style("database", "yellow"),
            style("src connect ok", "yellow"),
            style("dst connect ok", "yellow"),
        ]
    ]

    results.sort(key=lambda d: d["db"])

    failed_connection_exists = False
    for r in results:
        table.append(
            [
                style(r["db"], "green"),
                style(r["src"], "green" if r["src"] else "red"),
                style(r["dst"], "green" if r["dst"] else "red"),
            ]
        )
        # If any of the connections have failed in this DB, and the flag hasn't been set, set it.
        if not failed_connection_exists and (r["src"] is False or r["dst"] is False):
            failed_connection_exists = True

    echo(tabulate(table, headers="firstrow"))

    if failed_connection_exists:
        exit(1)


@run_with_configs(results_callback=_print_connectivity_results)
async def check_connectivity(config_future: Awaitable[DbupgradeConfig]) -> None:
    """
    Returns exit code 0 if pgbelt can connect to all databases in a datacenter
    (if db is not specified), or to both src and dst of a database.

    This is done by checking network access to the database ports ONLY.

    If any connection times out, the command will exit 1. It will test ALL connections
    before returning exit code 1 or 0, and output which connections passed/failed.
    """

    conf = await config_future

    src_future = open_connection(conf.src.ip, conf.src.port)
    src_logger = get_logger(conf.db, conf.dc, "connect.src")
    dst_future = open_connection(conf.dst.ip, conf.dst.port)
    dst_logger = get_logger(conf.db, conf.dc, "connect.dst")
    src_connect_ok = False
    dst_connect_ok = False

    # Source Connection Checks
    try:
        src_logger.info("Checking network access to port...")

        # Wait for 3 seconds, then raise TimeoutError
        _, writer = await wait_for(src_future, timeout=3)
        src_logger.debug("Can access network port.")
        writer.close()
        await writer.wait_closed()
        src_connect_ok = True
    except TimeoutError:
        src_logger.error("Cannot access network port. timed out.")
    except socket.gaierror as e:
        src_logger.error(f"Socket.gaierror {e}")
    except ConnectionRefusedError as e:
        src_logger.error(f"ConnectionRefusedError {e}")

    # Destination Connection Checks
    try:
        dst_logger.info("Checking network access to port...")

        # Wait for 3 seconds, then raise TimeoutError
        _, writer = await wait_for(dst_future, timeout=3)
        dst_logger.debug("Can access network port.")
        writer.close()
        await writer.wait_closed()
        dst_connect_ok = True
    except TimeoutError:
        dst_logger.error("Cannot access network port. timed out.")
    except socket.gaierror as e:
        dst_logger.error(f"Socket.gaierror {e}")
    except ConnectionRefusedError as e:
        dst_logger.error(f"ConnectionRefusedError {e}")

    # TODO: Exit code AFTER all have run

    return {"db": conf.db, "src": src_connect_ok, "dst": dst_connect_ok}


COMMANDS = [src_dsn, dst_dsn, check_pkeys, check_connectivity]



import asyncio
from logging import Logger
from os.path import join
from pgbelt.config.models import DbupgradeConfig
from pgbelt.util.asyncfuncs import isfile
from pgbelt.util.asyncfuncs import listdir
from pgbelt.util.asyncfuncs import makedirs
from pgbelt.util.postgres import table_empty
from re import search

from aiofiles import open as aopen
from asyncpg import create_pool

RAW = "schema"
NO_INVALID_NO_INDEX = "no_invalid_constraints_no_indexes"
ONLY_INVALID = "invalid_constraints"
ONLY_INDEXES = "indexes"


def schema_dir(db: str, dc: str) -> str:
    return f"schemas/{dc}/{db}"


def schema_file(db: str, dc: str, name: str) -> str:
    return join(schema_dir(db, dc), f"{name}.sql")


def table_dir(db: str, dc: str) -> str:
    return f"tables/{dc}/{db}"


def table_file(db: str, dc: str, name: str) -> str:
    return join(table_dir(db, dc), f"{name}.sql")


def _parse_dump_commands(out: str) -> list[str]:
    """
    Given a string containing output from pg_dump, return a list of strings where
    each is a complete postgres command. Commands may be multi-line.
    """
    lines = out.split("\n")
    commands = []

    for line in lines:
        stripped = line.strip()
        # if the line is whitespace only or a comment then ignore it
        if not stripped or stripped.startswith("--"):
            continue

        # if the last command is terminated or we don't have any yet start a new one
        if not commands or commands[-1].endswith(";\n"):
            commands.append(line + "\n")
        # otherwise we append to the last command because it must be multi-line
        else:
            commands[-1] += line + "\n"

    return commands


async def _execute_subprocess(
    command: list[str], finished_log: str, logger: Logger
) -> bytes:
    p = await asyncio.create_subprocess_exec(
        command[0],
        *command[1:],
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    out, err = await p.communicate()

    if p.returncode != 0:
        raise Exception(
            f"Couldn't do {command}, got code {p.returncode}.\n  out: {out.decode('utf-8')}\n  err: {err.decode('utf-8')}"
        )
    logger.debug(finished_log)
    return out


async def _dump_table(config: DbupgradeConfig, table: str, logger: Logger) -> None:
    """
    Dump a single table using pg_dump, strip unwanted lines, and save to file.
    """
    command = [
        "pg_dump",
        "--data-only",
        f'--table={config.schema_name}."{table}"',
        config.src.pglogical_dsn,
    ]

    out = await _execute_subprocess(command, f"dumped {table}", logger)
    content = out.decode("utf-8")

    # Strip out unwanted lines, stupid PG17
    keywords = [
        "transaction_timeout",
        # "SET statement_timeout", # This one is fine
        # "SET lock_timeout", # This one is fine
        # "SET idle_in_transaction_session_timeout", # This one is fine
        "SET client_encoding",
        "SET standard_conforming_strings",
        "SET check_function_bodies",
        "SET xmloption",
        "SET client_min_messages",
        "SET row_security",
        "pg_catalog.set_config",  # Stupid search path, this should not be run.
        "\\restrict",
        "\\unrestrict",
    ]
    lines = content.split("\n")
    filtered_lines = [
        line for line in lines if not any(keyword in line for keyword in keywords)
    ]
    filtered_content = "\n".join(filtered_lines)

    # Write the filtered content to file
    async with aopen(table_file(config.db, config.dc, table), "w") as f:
        await f.write(filtered_content)


async def dump_source_tables(
    config: DbupgradeConfig, tables: list[str], logger: Logger
) -> None:
    try:
        await makedirs(table_dir(config.db, config.dc))
    except FileExistsError:
        pass

    logger.info(f"Dumping tables {tables}")

    dumps = []
    for table in tables:
        dumps.append(_dump_table(config, table, logger))

    await asyncio.gather(*dumps)


async def load_dumped_tables(
    config: DbupgradeConfig, tables: list[str], logger: Logger
) -> None:
    # unless we get an explicit list of tables to load just load all the dump files
    if not tables:
        tables_dir = table_dir(config.db, config.dc)
        tables = [
            f.split(".")[0]
            for f in await listdir(tables_dir)
            if await isfile(join(tables_dir, f))
        ]

    logger.info(f"Loading dumped tables {tables}")

    # only load a dump file if the target table is completely empty
    async with create_pool(config.dst.root_uri, min_size=1) as pool:
        to_load = []
        for t in tables:
            if await table_empty(pool, t, config.schema_name, logger):
                to_load.append(table_file(config.db, config.dc, t))
            else:
                logger.warning(
                    f"Not loading {t}, table not empty. If this is unexpected please investigate."
                )

    loads = []
    for file in to_load:
        loads.append(
            _execute_subprocess(
                [
                    "psql",
                    config.dst.owner_dsn,
                    "-f",
                    file,
                ],
                f"loaded {file}",
                logger,
            )
        )

    await asyncio.gather(*loads)


async def dump_source_schema(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Dump the schema from the source db and write a file with the complete schema,
    one with only the CREATE INDEX statements from the schema,
    one with only the NOT VALID constraints from the schema,
    and one with everything but the NOT VALID constraints and the CREATE INDEX statements.
    """
    logger.info("Dumping schema...")

    command = [
        "pg_dump",
        "--schema-only",
        "--no-owner",
        "-n",
        config.schema_name,
        config.src.pglogical_dsn,
    ]

    # TODO: We should exclude the creation of a schema in the schema dump and load, and made that the responsibility of the user.
    # Confirm if the CREATE SCHEMA statement is included in the schema dump, and if yes, exclude it.
    # This will reveal itself in the integration test.

    out = await _execute_subprocess(command, "Retrieved source schema", logger)

    commands_raw = _parse_dump_commands(out.decode("utf-8"))

    commands = []
    for c in commands_raw:
        if "EXTENSION " not in c and "GRANT " not in c and "REVOKE " not in c:
            commands.append(c)

    try:
        await makedirs(schema_dir(config.db, config.dc))
    except FileExistsError:
        pass

    async with aopen(schema_file(config.db, config.dc, RAW), "w") as out:
        for command in commands:
            await out.write(command)

    async with aopen(schema_file(config.db, config.dc, ONLY_INVALID), "w") as out:
        for command in commands:
            if "NOT VALID" in command:
                await out.write(command)

    async with aopen(schema_file(config.db, config.dc, ONLY_INDEXES), "w") as out:
        for command in commands:
            if "CREATE" in command and "INDEX" in command:
                await out.write(command)

    async with aopen(
        schema_file(config.db, config.dc, NO_INVALID_NO_INDEX), "w"
    ) as out:
        for command in commands:
            if "NOT VALID" not in command and not (
                "CREATE" in command and "INDEX" in command
            ):
                await out.write(command)

    logger.debug("Finished dumping schema.")


async def apply_target_schema(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Load the schema dumped from the source into the target excluding NOT VALID constraints and CREATE INDEX statements.
    """
    logger.info("Loading schema without constraints...")

    command = [
        "psql",
        config.dst.owner_dsn,
        "-f",
        schema_file(config.db, config.dc, NO_INVALID_NO_INDEX),
    ]

    await _execute_subprocess(command, "Finished loading schema.", logger)


async def dump_dst_not_valid_constraints(
    config: DbupgradeConfig, logger: Logger
) -> None:
    """
    Dump NOT VALID Constraints from the target database.
    Used when schema is loaded in outside of pgbelt.
    """

    logger.info("Dumping target NOT VALID constraints...")

    command = [
        "pg_dump",
        "--schema-only",
        "--no-owner",
        "-n",
        config.schema_name,
        config.dst.pglogical_dsn,
    ]

    out = await _execute_subprocess(command, "Retrieved target schema", logger)

    # No username replacement needs to be done, so replace dst user with the same.
    commands_raw = _parse_dump_commands(
        out.decode("utf-8"), config.dst.owner_user.name, config.dst.owner_user.name
    )

    commands = []
    for c in commands_raw:
        if "NOT VALID" in command:
            if config.tables:
                regex_matches = search(
                    r"ALTER TABLE [ONLY ]*(?P<table>[a-zA-Z0-9._]+)+\s+ADD CONSTRAINT (?P<constraint>[a-zA-Z0-9._]+)+.*",
                    c,
                )
                if not regex_matches:
                    continue
                table = regex_matches.groupdict()["table"]
                if config.tables and table in config.tables:
                    commands.append(c)
            else:
                commands.append(c)

    try:
        await makedirs(schema_dir(config.db, config.dc))
    except FileExistsError:
        pass

    async with aopen(schema_file(config.db, config.dc, ONLY_INVALID), "w") as out:
        for command in commands:
            await out.write(command)

    logger.debug("Finished dumping NOT VALID constraints from the target.")


async def remove_dst_not_valid_constraints(
    config: DbupgradeConfig, logger: Logger
) -> None:
    """
    Remove the NOT VALID constraints from the schema of the target database.
    Only use if target schema was loaded in without pgbelt.
    """
    logger.info("Looking for previously dumped NOT VALID constraints...")

    async with aopen(schema_file(config.db, config.dc, ONLY_INVALID), "r") as f:
        not_valid_constraints = await f.read()

    logger.info("Removing NOT VALID constraints from the target...")

    queries = ""
    for c in not_valid_constraints.split(";"):
        regex_matches = search(
            r"ALTER TABLE [ONLY ]*(?P<table>[a-zA-Z0-9._]+)+\s+ADD CONSTRAINT (?P<constraint>[a-zA-Z0-9._]+)+.*",
            c,
        )
        if not regex_matches:
            continue
        table = regex_matches.groupdict()["table"]
        constraint = regex_matches.groupdict()["constraint"]

        if (config.tables and table in config.tables) or not config.tables:
            queries = queries + f"ALTER TABLE {table} DROP CONSTRAINT {constraint};"

    if queries != "":
        command = ["psql", config.dst.owner_dsn, "-c", queries]

        await _execute_subprocess(
            command, "Finished removing NOT VALID constraints from the target.", logger
        )
    else:
        logger.info("No NOT VALID detected for removal.")


async def apply_target_constraints(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Load the NOT VALID constraints that were excluded from the schema. Should be called after replication during
    downtime before allowing writes into the target.
    """
    logger.info("Loading NOT VALID constraints...")

    command = [
        "psql",
        config.dst.owner_dsn,
        "-f",
        schema_file(config.db, config.dc, ONLY_INVALID),
    ]

    await _execute_subprocess(
        command, "Finished loading NOT VALID constraints.", logger
    )


async def remove_dst_indexes(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Remove the INDEXes from the schema of the target database.
    Only use if target schema was loaded in without pgbelt.
    """
    logger.info("Looking for previously dumped CREATE INDEX statements...")

    async with aopen(schema_file(config.db, config.dc, ONLY_INDEXES), "r") as f:
        create_index_statements = await f.read()

    logger.info("Removing Indexes from the target...")

    for c in create_index_statements.split(";"):
        regex_matches = search(
            r"CREATE [UNIQUE ]*INDEX (?P<index>[a-zA-Z0-9._]+)+.*",
            c,
        )
        if not regex_matches:
            continue
        index = regex_matches.groupdict()["index"]
        if config.schema_name:
            index = f"{config.schema_name}.{index}"

        # DROP the index
        # Note that the host DSN must have a statement timeout of 0.
        # Example DSN: `host=server-hostname user=user dbname=db_name options='-c statement_timeout=3600000'`
        host_dsn = config.dst.owner_dsn + " options='-c statement_timeout=0'"

        # DROP INDEX IF EXISTS so no need to catch exceptions
        command = ["psql", host_dsn, "-c", f"DROP INDEX IF EXISTS {index};"]
        logger.info(f"Dropping index {index} on the target...")
        await _execute_subprocess(
            command, f"Finished dropping index {index} on the target.", logger
        )


async def create_target_indexes(
    config: DbupgradeConfig, logger: Logger, during_sync=False
) -> None:
    """
    Create indexes on the target that were excluded from the schema during setup.
    Should be called once bulk syncing is complete, and before cutover.

    Runs in serial for now with this async code.
    TODO: make this run in parallel (beware risk of building too many indexes at once, resource heavy)
    """

    if during_sync:
        logger.warning(
            "Attempting to create indexes on the target. If indexes were not created before the cutover window, this can take a long time."
        )

    logger.info("Looking for previously dumped CREATE INDEX statements...")

    async with aopen(schema_file(config.db, config.dc, ONLY_INDEXES), "r") as f:
        create_index_statements = await f.read()

    logger.info("Creating indexes on the target...")

    for c in create_index_statements.split(";"):
        # Get the Index Name
        regex_matches = search(
            r"CREATE [UNIQUE ]*INDEX (?P<index>[a-zA-Z0-9._\"]+)+.*",
            c,
        )
        if not regex_matches:
            continue
        index = regex_matches.groupdict()["index"]

        # Sometimes the index name is quoted, so remove the quotes
        index = index.replace('"', "")

        # Create the index
        # Note that the host DSN must have a statement timeout of 0.
        # Example DSN: `host=server-hostname user=user dbname=db_name options='-c statement_timeout=3600000'`
        host_dsn = config.dst.owner_dsn + " options='-c statement_timeout=0'"
        command = ["psql", host_dsn, "-c", f"{c};"]
        logger.info(f"Creating index {index} on the target...")
        try:
            await _execute_subprocess(
                command, f"Finished creating index {index} on the target.", logger
            )
        except Exception as e:
            if f'relation "{index}" already exists' in str(e):
                logger.info(f"Index {index} already exist on the target.")
            else:
                raise Exception(e)



from asyncio import gather
from collections.abc import Awaitable
from logging import Logger

from asyncpg import create_pool
from asyncpg import Pool
from pgbelt.cmd.helpers import run_with_configs
from pgbelt.config.models import DbupgradeConfig
from pgbelt.util.dump import apply_target_constraints
from pgbelt.util.dump import create_target_indexes
from pgbelt.util.dump import dump_source_tables
from pgbelt.util.dump import load_dumped_tables
from pgbelt.util.logs import get_logger
from pgbelt.util.postgres import analyze_table_pkeys
from pgbelt.util.postgres import compare_100_random_rows
from pgbelt.util.postgres import compare_latest_100_rows
from pgbelt.util.postgres import compare_tables_without_pkeys
from pgbelt.util.postgres import dump_sequences
from pgbelt.util.postgres import load_sequences
from pgbelt.util.postgres import run_analyze
from typer import Option


async def _sync_sequences(
    targeted_sequences: list[str],
    schema: str,
    src_pool: Pool,
    dst_pool: Pool,
    src_logger: Logger,
    dst_logger: Logger,
) -> None:

    seq_vals = await dump_sequences(src_pool, targeted_sequences, schema, src_logger)
    await load_sequences(dst_pool, seq_vals, schema, dst_logger)


@run_with_configs
async def sync_sequences(config_future: Awaitable[DbupgradeConfig]) -> None:
    """
    Retrieve the current value of all sequences in the source database and update
    the sequences in the target to match.
    """
    conf = await config_future
    pools = await gather(
        create_pool(conf.src.pglogical_uri, min_size=1),
        create_pool(conf.dst.root_uri, min_size=1),
    )
    src_pool, dst_pool = pools
    try:
        src_logger = get_logger(conf.db, conf.dc, "sync.src")
        dst_logger = get_logger(conf.db, conf.dc, "sync.dst")
        await _sync_sequences(
            conf.sequences, conf.schema_name, src_pool, dst_pool, src_logger, dst_logger
        )
    finally:
        await gather(*[p.close() for p in pools])


@run_with_configs(skip_dst=True)
async def dump_tables(
    config_future: Awaitable[DbupgradeConfig],
    tables: list[str] = Option([], help="Specific tables to dump"),
) -> None:
    """
    Dump all tables without primary keys from the source database and save
    them to files locally.

    You may also provide a list of tables to dump with the
    --tables option and only these tables will be dumped.
    """
    conf = await config_future
    logger = get_logger(conf.db, conf.dc, "sync.src")

    if tables:
        tables = tables.split(",")
    else:
        async with create_pool(conf.src.pglogical_uri, min_size=1) as src_pool:
            _, tables, _ = await analyze_table_pkeys(src_pool, conf.schema_name, logger)

        if conf.tables:
            tables = [t for t in tables if t in conf.tables]

    await dump_source_tables(conf, tables, logger)


@run_with_configs(skip_src=True)
async def load_tables(
    config_future: Awaitable[DbupgradeConfig],
    tables: list[str] = Option([], help="Specific tables to load"),
):
    """
    Load all locally saved table data files into the destination db. A table will
    only be loaded into the destination if it currently contains no rows.

    You may also provide a list of tables to load with the
    --tables option and only these files will be loaded.
    """
    conf = await config_future
    logger = get_logger(conf.db, conf.dc, "sync.dst")

    if tables:
        tables = tables.split(",")
    else:
        if conf.tables:
            tables = [t for t in tables if t in conf.tables]
        else:
            tables = []

    await load_dumped_tables(conf, tables, logger)


@run_with_configs(skip_src=True)
async def analyze(config_future: Awaitable[DbupgradeConfig]) -> None:
    """
    Run ANALYZE in the destination database. This should be run after data is
    completely replicated and before applications are allowed to use the new db.
    """
    conf = await config_future
    logger = get_logger(conf.db, conf.dc, "sync.dst")
    async with create_pool(
        conf.dst.root_uri,
        min_size=1,
        server_settings={
            "statement_timeout": "0",
        },
    ) as dst_pool:
        await run_analyze(dst_pool, logger)


@run_with_configs
async def validate_data(config_future: Awaitable[DbupgradeConfig]) -> None:
    """
    Compares data in the source and target databases. Both a random sample and a
    sample of the latest rows will be compared for each table. Does not validate
    the entire data set.
    """
    conf = await config_future
    pools = await gather(
        create_pool(conf.src.pglogical_uri, min_size=1),
        create_pool(conf.dst.owner_uri, min_size=1),
    )
    src_pool, dst_pool = pools

    try:
        logger = get_logger(conf.db, conf.dc, "sync")
        await gather(
            compare_100_random_rows(
                src_pool, dst_pool, conf.tables, conf.schema_name, logger
            ),
            compare_latest_100_rows(
                src_pool, dst_pool, conf.tables, conf.schema_name, logger
            ),
            compare_tables_without_pkeys(
                src_pool, dst_pool, conf.tables, conf.schema_name, logger
            ),
        )
    finally:
        await gather(*[p.close() for p in pools])


async def _dump_and_load_all_tables(
    conf: DbupgradeConfig, src_pool: Pool, src_logger: Logger, dst_logger: Logger
) -> None:
    _, tables, _ = await analyze_table_pkeys(src_pool, conf.schema_name, src_logger)
    if conf.tables:
        tables = [t for t in tables if t in conf.tables]
    await dump_source_tables(conf, tables, src_logger)
    await load_dumped_tables(conf, tables, dst_logger)


@run_with_configs
async def sync(
    config_future: Awaitable[DbupgradeConfig], no_schema: bool = False
) -> None:
    """
    Sync and validate all data that is not replicated with pglogical. This includes all
    tables without primary keys and all sequences. Also loads any previously omitted
    NOT VALID constraints into the destination db and runs ANALYZE in the destination.

    This command is equivalent to running the following commands in order:
    sync-sequences, sync-tables, validate-data, load-constraints, analyze.
    Though here they may run concurrently when possible.
    """
    conf = await config_future
    pools = await gather(
        create_pool(conf.src.pglogical_uri, min_size=1),
        create_pool(conf.dst.root_uri, min_size=1),
        create_pool(conf.dst.owner_uri, min_size=1),
        create_pool(
            conf.dst.root_uri,
            min_size=1,
            server_settings={
                "statement_timeout": "0",
            },
        ),
    )
    src_pool, dst_root_pool, dst_owner_pool, dst_root_no_timeout_pool = pools

    try:
        src_logger = get_logger(conf.db, conf.dc, "sync.src")
        dst_logger = get_logger(conf.db, conf.dc, "sync.dst")
        validation_logger = get_logger(conf.db, conf.dc, "sync")

        await gather(
            _sync_sequences(
                conf.sequences,
                conf.schema_name,
                src_pool,
                dst_root_pool,
                src_logger,
                dst_logger,
            ),
            _dump_and_load_all_tables(conf, src_pool, src_logger, dst_logger),
        )

        # Creating indexes should run before validations and ANALYZE, but after all the data exists
        # in the destination database.

        # Do not load NOT VALID constraints or create INDEXes for exodus-style migrations
        if not no_schema:
            await gather(
                apply_target_constraints(conf, dst_logger),
                create_target_indexes(conf, dst_logger, during_sync=True),
            )

        await gather(
            compare_100_random_rows(
                src_pool,
                dst_owner_pool,
                conf.tables,
                conf.schema_name,
                validation_logger,
            ),
            compare_latest_100_rows(
                src_pool,
                dst_owner_pool,
                conf.tables,
                conf.schema_name,
                validation_logger,
            ),
            run_analyze(dst_root_no_timeout_pool, dst_logger),
        )
    finally:
        await gather(*[p.close() for p in pools])


COMMANDS = [
    sync_sequences,
    dump_tables,
    load_tables,
    analyze,
    validate_data,
    sync,
]



from asyncio import gather
from asyncio import run
from collections.abc import Awaitable
from collections.abc import Callable
from functools import wraps
from inspect import iscoroutinefunction
from inspect import Parameter
from inspect import signature
from typing import Any
from typing import Optional  # noqa: F401 # Needed until tiangolo/typer#522 is fixed)
from typing import TypeVar

from pgbelt.config import get_all_configs_async
from pgbelt.config import get_config_async
from typer import Argument
from typer import Typer


T = TypeVar("T")


def run_with_configs(
    decorated_func: Callable[..., Awaitable[Optional[T]]] = None,
    skip_src: bool = False,
    skip_dst: bool = False,
    results_callback: Optional[Callable[[list[T]], Awaitable[Optional[Any]]]] = None,
) -> Callable:
    """
    Decorator for async commands. Implementations should take one Awaitable[DbupgradeConfig] arg
    and do some operation on the databases in it. This wrapper handles looking up the
    config and executing the command. The decorated result can be run either on one db only
    or on the entire datacenter concurrently.

    You may also provide a callback that will be called on the results of the command. Useful
    for displaying the output of interrogative commands.
    """

    def decorator(func):
        if skip_src and skip_dst:
            func.__doc__ += (
                "\n\n    Can be run with both src and dst set null in the config file."
            )
        elif skip_src:
            func.__doc__ += "\n\n    Can be run with a null src in the config file."
        elif skip_dst:
            func.__doc__ += "\n\n    Can be run with a null dst in the config file."
        else:
            func.__doc__ += (
                "\n\n    Requires both src and dst to be not null in the config file."
            )

        # The name, docstring, and signature of the implementation is preserved. Important for add_command
        @wraps(func)
        async def wrapper(dc: str, db: Optional[str], **kwargs):
            # If db is specified we only want to run on one of them
            if db is not None:
                results = [
                    await func(
                        get_config_async(db, dc, skip_src=skip_src, skip_dst=skip_dst),
                        **kwargs
                    )
                ]
            else:
                # if the db is not provided run on all the dbs in the dc
                results = await gather(
                    *[
                        func(fut, **kwargs)
                        async for fut in get_all_configs_async(
                            dc, skip_src=skip_src, skip_dst=skip_dst
                        )
                    ]
                )

            # Call the callback if provided.
            if results_callback is None:
                return results
            return await results_callback(results)

        return wrapper

    # makes either @decorator or @decorator(...) work to decorate a function
    if decorated_func is None:
        return decorator
    return decorator(decorated_func)


def add_command(app: Typer, command: Callable):
    """
    Helper which attaches a function to the given typer app. Merges
    the signature of the underlying implementation with standard arguments.
    This allows command options to be defined on the implementation and all
    commands in belt to share common arguments defined in one place.

    If a command is async, then it may be run on all dbs in a dc concurrently.
    Otherwise we assume it only makes sense to run it on one.
    """
    # Give typer the name of the actual implementing function
    name = command.__name__.replace("_", "-")

    # If async assume command can be run on a whole datacenter and make db optional
    if iscoroutinefunction(command):

        @app.command(name=name)
        def cmdwrapper(dc: str, db: Optional[str] = Argument(None), **kwargs):
            run(command(dc, db, **kwargs))

    # Synchronous commands can only be run on one db at a time
    else:

        @app.command(name=name)
        def cmdwrapper(dc: str, db: str, **kwargs):
            command(db, dc, **kwargs)

    # remove the **kwargs since typer doesn't do anything with it
    wrap_signature = signature(cmdwrapper)
    wrap_params = wrap_signature.parameters.copy()
    wrap_params.popitem()

    # Remove any args without defaults from the implementation's signature
    # so we are left with only what typer interprets as options
    cmd_params = signature(command).parameters.copy()
    cmd_params_copy = cmd_params.copy()
    while (
        cmd_params_copy
        and cmd_params_copy.popitem(last=False)[1].default is Parameter.empty
    ):
        cmd_params.popitem(last=False)

    # merge the arguments from the wrapper with the options from the implementation
    wrap_params.update(cmd_params)

    # set the signature for typer to read
    cmdwrapper.__signature__ = wrap_signature.replace(parameters=wrap_params.values())

    # the docstring on the implementation will be used as typer help
    cmdwrapper.__doc__ = command.__doc__
    if iscoroutinefunction(command):
        cmdwrapper.__doc__ += (
            "\n\n    If the db name is not given run on all dbs in the dc."
        )



from importlib import import_module
from json import JSONDecodeError
from logging import Logger
from typing import Optional  # noqa: F401 # Needed until tiangolo/typer#522 is fixed)

from aiofiles import open as aopen
from pgbelt.config.models import DbupgradeConfig
from pgbelt.util.logs import get_logger
from pydantic import BaseModel
from pydantic import ValidationError


def remote_conf_path(db: str, dc: str) -> str:
    return f"remote-configs/{dc}/{db}/config.json"


class RemoteConfigError(Exception):
    """
    raised by the resolve method of a resolver when it could not retrieve
    configuration due to an error.
    """

    pass


class RemoteConfigDefinition(BaseModel):
    """
    Pydantic model representing the contents of a remote-config json
    before we have the actual resolver to put it in. So the only required
    key is `resolver_path`. The rest are passed into the resolver model.
    """

    resolver_path: str

    class Config:
        extra = "allow"


class BaseResolver(BaseModel):
    """
    Remote configuration resolvers must subclass this.

    db: str the database name
    dc: str the datacenter name
    skip_src: bool don't retrieve the source db configuration
    skip_dst: bool don't retrieve the destination db configuration
    logger: Logger your resolver should log through this logger

    Your resolver subclass will be a pydantic model. Any attributes you define
    other than the ones already in here will come from your remote config files.
    """

    db: str
    dc: str
    skip_src: bool
    skip_dst: bool
    logger: Logger

    class Config:
        arbitrary_types_allowed = True
        extra = "ignore"

    async def resolve(self) -> Optional[DbupgradeConfig]:
        """
        Called to retrieve the configuration from wherever your resolver gets it.
        Return configuration as a DbupgradeConfig.

        If you should have been able to get configuration but encountered an error
        then throw a RemoteConfigError. If there was no config and no error return None
        """
        raise NotImplementedError


async def load_remote_conf_def(
    config_file: str, logger: Logger
) -> Optional[RemoteConfigDefinition]:
    try:
        logger.debug(f"Reading remote config definition from file {config_file}")
        async with aopen(config_file, mode="r") as f:
            raw_json = await f.read()

        return RemoteConfigDefinition.model_validate_json(raw_json)
    except FileNotFoundError:
        logger.error(f"No remote config definition exists at {config_file}")
    except JSONDecodeError:
        logger.error(f"Remote config definition in {config_file} was malformed JSON.")
    except ValidationError:
        logger.error(f"Remote config definition in {config_file} was not valid.")


async def resolve_remote_config(
    db: str, dc: str, skip_src: bool = False, skip_dst: bool = False
) -> Optional[DbupgradeConfig]:
    """
    Loads the referenced remote configuration json file, tries to import the
    specified resolver class, executes its resolve method, and returns the
    resulting DbupgradeConfig or None if there was an error
    """

    # set up the logger
    logger = get_logger(db, dc, "remote-config")

    # load the remote config from the json file
    definition = await load_remote_conf_def(remote_conf_path(db, dc), logger)

    if definition is None:
        return None

    module, classname = definition.resolver_path.rsplit(".", 1)

    try:
        resolver_module = import_module(module)
    except ModuleNotFoundError:
        logger.error(f"Could not find module {module}")
        return None

    try:
        resolver_class = getattr(resolver_module, classname)
    except AttributeError:
        logger.error(f"Config resolver class {classname} does not exist in {module}")
        return None

    if not issubclass(resolver_class, BaseResolver):
        logger.error(
            f"Config resolver class {classname} from {module} is not a config resolver"
        )
        return None

    try:
        resolver_dict = definition.dict()

        resolver_dict.update(
            {
                "db": db,
                "dc": dc,
                "skip_src": skip_src,
                "skip_dst": skip_dst,
                "logger": logger.getChild(classname),
            }
        )

        resolver = resolver_class(**resolver_dict)
    except ValidationError:
        logger.error(
            f"Remote config definition for {db} {dc} was not valid for {resolver_class.__name__}"
        )
        return None

    try:
        config = await resolver.resolve()
    except NotImplementedError:
        logger.error(
            f"Config resolver class {classname} from {module} does not implement resolve"
        )
        return None
    except RemoteConfigError as e:
        logger.error(
            f"Failed to resolve remote configuration for {db} {dc}. RemoteConfigError {e}"
        )
        return None
    except ValidationError:
        logger.error(
            f"Configuration for {db} {dc} resolved by {resolver_class.__name__} was not a valid DbupgradeConfig"
        )
        return None

    logger.info(
        f"Successfully resolved remote configuration for {db} {dc} using {resolver_class.__name__}"
    )

    return config
